--------------------------------------------------------------------------------
--------------------------------------------------------------------------------

function gadget:GetInfo()
  return {
    name      = "UnitRetreat",
    desc      = "Adds a button that sets the health % " ..
                "under which units will retreat",
    author    = "quantum",
    date      = "June 30, 2007",
    license   = "GNU GPL, v2 or later",
    layer     = 0,
    enabled   = true  --  loaded by default?
  }
end

--------------------------------------------------------------------------------
--------------------------------------------------------------------------------

if (not gadgetHandler:IsSyncedCode()) then
  return false  --  silent removal
end

--------------------------------------------------------------------------------
--------------------------------------------------------------------------------

--[[
  FIXME
  - rename "Do Or Die" to "No Retreat" or "Retreat Off" ? check
  - better retreat unit tracking  (faster than the tag method)
  - retreat indicator  (and not the cmdDesc disables  ;-)
  - more retreat % levels, or try something new (initially 50%) ?:
      LMB = on/off, RMB = ++, RMB+SHIFT = --?, CTRL for 10% / 20% steps
  - stagger the unit checks
--]]


--Speed-ups

local GetUnitPosition = Spring.GetUnitPosition
local GetUnitDefID    = Spring.GetUnitDefID
local GetUnitHealth   = Spring.GetUnitHealth
local GiveOrderToUnit = Spring.GiveOrderToUnit
local GetUnitCommands = Spring.GetUnitCommands
local FindUnitCmdDesc = Spring.FindUnitCmdDesc
local insert          = table.insert
local sort            = table.sort
local rand            = math.random
local s               = Spring.SendMessage


--------------------------------------------------------------------------------
--------------------------------------------------------------------------------

CMD_RETREAT = 32200

local havens = {}

local retreatTags = {}    -- active retreat move commands

local retreatOrders = {}  -- cmdDesc retreat states

local havenDefs = {
  armcom = true,
  corcom = true,
  armnanotc = true,
  cornanotc = true,
}

local retreatCmdDesc = {
  id      = CMD_RETREAT,
  type    = CMDTYPE.ICON_MODE,
  name    = 'Retreat',
  cursor  = 'Retreat',
  action  = 'retreat',
  tooltip = 'Orders: never retreat',
  params  = { '0', 'Retreat Off', 'Retreat 30%', 'Retreat 60%' }
}
  
local dist = 160
local maxDistSqr = dist * dist


--------------------------------------------------------------------------------
--------------------------------------------------------------------------------

local function AddRetreatCmdDesc(unitID)
  if (FindUnitCmdDesc(unitID, CMD_RETREAT)) then
    return  -- already exists
  end
  local insertID = 
    FindUnitCmdDesc(unitID, CMD.CLOAK)      or
    FindUnitCmdDesc(unitID, CMD.ONOFF)      or
    FindUnitCmdDesc(unitID, CMD.TRAJECTORY) or
    FindUnitCmdDesc(unitID, CMD.REPEAT)     or
    FindUnitCmdDesc(unitID, CMD.MOVE_STATE) or
    FindUnitCmdDesc(unitID, CMD.FIRE_STATE) or
    123456 -- back of the pack
  retreatCmdDesc.params[1] = '0'
  Spring.InsertUnitCmdDesc(unitID, insertID + 1, retreatCmdDesc)
end


local function UpdateButton(unitID, statusStr)
  local cmdDescID = FindUnitCmdDesc(unitID, CMD_RETREAT)
  if (cmdDescID == nil) then
    return
  end

  local tooltip
  if (statusStr == '1') then
    tooltip = 'Orders: retreat at less than 30% of hit points remaining.'
  elseif (statusStr == '2') then
    tooltip = 'Orders: retreat at less than 60% of hit points remaining.'
  else
    tooltip = 'Orders: never retreat.'
  end

  retreatCmdDesc.params[1] = statusStr

  Spring.EditUnitCmdDesc(unitID, cmdDescID, { 
    params  = retreatCmdDesc.params, 
    tooltip = tooltip,
  })
end


local function CanReallyMove(unitDefID)
  -- let factories have the retreat button, but don't check them
  -- (they do have canMove for new units, but their speed should be 0)
  local ud = UnitDefs[unitDefID]
  if (ud == nil) then
    return false
  end
  return (ud.speed > 0)
end


local function RetreatCommand(unitID, unitDefID, cmdParams, teamID)
  local status
  if cmdParams[1] == 1 then
    
    retreatOrders[unitID] = { 1, teamID, CanReallyMove(unitDefID) }
    status = '1'
  elseif cmdParams[1] == 2 then
    retreatOrders[unitID] = { 2, teamID, CanReallyMove(unitDefID) }
    status = '2'
  else
    retreatOrders[unitID] = nil
    status = '0'
  end
  UpdateButton(unitID, status)
end


local function FindClosestHaven(unitID, teamID)
  local ux, _, uz = GetUnitPosition(unitID)
  local closestDistSqr = math.huge
  local cx, cy, cz  --  closest coordinates
  for havenUnitID, _ in pairs(havens[teamID]) do
    if (not havenUnitID) then return 0, 0, 0, 0 end
    local hx, hy, hz = GetUnitPosition(havenUnitID)
    local dSquared = (hx - ux)^2 + (hz - uz)^2
    if (dSquared < closestDistSqr) then
      closestDistSqr = dSquared
      cx = hx; cy = hy; cz = hz
    end
  end
  if (not cx) then return 0, 0, 0, 0 end  -- should not happen
  local px = cx + dist - rand(0, dist*2)
  local py = cy
  local pz = cz + dist - rand(0, dist*2)
  return px, py, pz, closestDistSqr
end


local function IsRetreating(unitID)
  local tag = retreatTags[unitID]
  if (tag == nil) then
    return false
  end
  local cmds = GetUnitCommands(unitID, 1)
  if ((cmds == nil) or (cmds[1] == nil)) then
    retreatTags[unitID] = nil  --  might as well clear it
    return false
  end
  if (cmds[1].tag == tag) then
    return true
  end
  retreatTags[unitID] = nil  --  might as well clear it
  return false
end


--------------------------------------------------------------------------------
--------------------------------------------------------------------------------

function gadget:UnitCreated(unitID, unitDefID, teamID, builderID)
  local ud = UnitDefs[unitDefID]
  if (havenDefs[ud.name]) then
    havens[teamID][unitID] = true
  elseif (ud.canMove) then
    AddRetreatCmdDesc(unitID)
    retreatOrders[unitID] = nil
    UpdateButton(unitID, '0')
    if (builderID) then
      local builderInfo = retreatOrders[builderID]
      if (builderInfo) then
        local builderState = builderInfo[1]
        RetreatCommand(unitID, unitDefID, { builderInfo[1] }, teamID)
      end
    end
  end
end


function gadget:UnitDestroyed(unitID, _, teamID)
  havens[teamID][unitID] = nil
  retreatOrders[unitID]  = nil
  retreatTags[unitID]    = nil
end


function gadget:UnitTaken(unitID, unitDefID, oldTeam, newTeam)
  havens[newTeam][unitID] = havens[oldTeam][unitID]
  havens[oldTeam][unitID] = nil
  retreatTags[unitID] = nil
  retreatOrders[unitID] = nil
  RetreatCommand(unitID, unitDefID, { 0 }, newTeam)
end


function gadget:Initialize()
  gadgetHandler:RegisterCMDID(CMD_RETREAT)
  local teamList = Spring.GetTeamList()
  for _, teamID in ipairs(teamList) do
    havens[teamID] = {}
  end
  for _, unitID in ipairs(Spring.GetAllUnits()) do
    local teamID = Spring.GetUnitTeam(unitID)
    local unitDefID = GetUnitDefID(unitID)
    gadget:UnitCreated(unitID, unitDefID, teamID)
  end
end


function gadget:Shutdown()
  for _, unitID in ipairs(Spring.GetAllUnits()) do
    local cmdDescID = FindUnitCmdDesc(unitID, CMD_RETREAT)
    if (cmdDescID) then
      Spring.RemoveUnitCmdDesc(unitID, cmdDescID)
    end
  end
end


function gadget:AllowCommand(unitID, unitDefID, teamID, cmdID, cmdParams, _)
  local returnvalue
  if cmdID ~= CMD_RETREAT then
    return true
  end
  RetreatCommand(unitID, unitDefID, cmdParams, teamID)  
  return false
end


function gadget:CommandFallback(unitID, unitDefID, teamID, cmdID, cmdParams, _)
  -- FIXME: RETREAT is a 'state' command, don't need to wait for the fallback
  --        (AllowCommand() has already culled it...)
  if cmdID ~= CMD_RETREAT then
    return false
  end
  RetreatCommand(unitID, unitDefID, cmdParams, teamID)
  return true, true
end


function gadget:GameFrame(n)
  if ((n % 33) < 0.1) then
    for unitID, list in pairs(retreatOrders) do
      if (list[3]) then
        local teamID = list[2]
        local health, maxHealth = GetUnitHealth(unitID)
        local ratio = health / maxHealth
        if ((((list[1] == 1) and (ratio < 0.3)) or
             ((list[1] == 2) and (ratio < 0.6)))
            and (#havens[teamID] > 0)) then
          if (not IsRetreating(unitID)) then
            local hx, hy, hz, dSquared = FindClosestHaven(unitID, teamID)
            if (dSquared > maxDistSqr) then
              -- using OPT_INTERNAL so that the CMD.MOVE order
              -- is not cycled when the unit has repeat enabled
              GiveOrderToUnit(unitID,
                              CMD.INSERT, -- push the order at front
                              { 0, CMD.MOVE, CMD.OPT_INTERNAL, hx, hy, hz},
                              CMD.OPT_ALT) -- ALT makes the 0 positional
              local cmds = Spring.GetUnitCommands(unitID, 1)
              if (cmds and cmds[1]) then
                retreatTags[unitID] = cmds[1].tag
              end
            end
          end
        end
      end
    end
  end
end  -- maybe I should apply the 'continue' lua patch   ;-)


--------------------------------------------------------------------------------
--------------------------------------------------------------------------------
